from itertools import chain

from admin_tools.db.fields import GenericForeignKey
from admin_tools.services.objects import copy_func
from admin_tools.widgets.autocomplete import ContentTypeAutocompleteSelect
from django import forms
from django.contrib.admin.checks import BaseModelAdminChecks
from django.contrib.admin.options import InlineModelAdmin
from django.contrib.admin.utils import flatten_fieldsets
from django.contrib.contenttypes.models import ContentType
from django.db.models import ForeignKey, ManyToManyField
from django.utils.translation import gettext_lazy as _


class AdminAutocompleteFieldsMixin:
    def get_autocomplete_fields(self, request):
        autocomplete_fields = super().get_autocomplete_fields(request)

        if autocomplete_fields:
            return autocomplete_fields

        admin_fields = self._get_admin_fields(request)
        relation_fields = self._get_relation_fields()

        if not admin_fields:
            return tuple(relation_fields)

        return tuple(set(admin_fields) & set(relation_fields))

    def _get_relation_fields(self):
        return (
            field.name
            for field in self.model._meta.get_fields()
            if self._is_relation_field(field)
        )

    def _get_admin_fields(self, request):
        if self.fields:
            return self.get_fields(request)
        elif self.fieldsets:
            return flatten_fieldsets(self.get_fieldsets(request))

    def _is_relation_field(self, field):
        return (isinstance(field, (ForeignKey, ManyToManyField))
                and bool(
                self.admin_site._registry.get(field.remote_field.model)))

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_autocomplete_mixin(),
        ]

    def _check_autocomplete_mixin(self):
        if 'autocomplete_fields' in self.__class__.__dict__.keys():
            return []

        check_autocomplete_fields_item = BaseModelAdminChecks()._check_autocomplete_fields_item

        return list(chain.from_iterable([
            check_autocomplete_fields_item(self, field_name,
                                           f'autocomplete_fields[{index:d}]')
            for index, field_name in
            enumerate(self.get_autocomplete_fields(None))
        ]))


class BaseGenericForeignKeyMixin:
    def get_fields(self, request, obj=None):
        fields = super().get_fields(request, obj=obj)
        if fields:
            fields = self._update_gfk_fields(fields)

        return fields

    def _update_gfk_fields(self, fields):
        gfk_fields = self._get_generic_foreign_keys(self.model)

        if not gfk_fields:
            return fields

        gfk_map = {f.name: f for f in gfk_fields}

        updated = []
        for field in fields:
            gfk = gfk_map.get(field)

            if not (gfk and isinstance(self, InlineModelAdmin)):
                updated = self._append_value(updated, field)

            if gfk:
                updated = self._append_value(updated, gfk.ct_field)
                updated = self._append_value(updated, gfk.fk_field)

        return tuple(updated)

    def _get_generic_foreign_keys(self, model):
        return [
            f
            for f in model._meta.get_fields()
            if isinstance(f, GenericForeignKey)
        ]

    def _append_value(self, fields, value):
        if value not in fields:
            fields.append(value)
        return fields

    def _set_validators_gfk(self, form):
        for field in self._get_generic_foreign_keys(self.model):
            _copy_func = copy_func(self._clean_formfield)
            _copy_func.__defaults__ = (field,)

            setattr(form, f'clean_{field.name}', _copy_func)

    def _clean_formfield(self, gfk_field):
        ct_field = self.cleaned_data.get(gfk_field.ct_field)
        fk_field = self.cleaned_data.get(gfk_field.fk_field)

        if ct_field and fk_field:
            return
        elif not ct_field and not fk_field:
            return

        msg = _('Both values must be filled or cleared')
        self.add_error(gfk_field.ct_field, msg)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        generic_foreign_keys = self._get_generic_foreign_keys(self.model)

        is_gfk, gfk = self._is_generic_foreign_key(
            db_field.name, generic_foreign_keys
        )

        form_field = super().formfield_for_foreignkey(
            db_field, request, **kwargs
        )

        if not is_gfk:
            return form_field

        attrs = self._get_content_type_autocomplete_attrs(gfk)
        db = kwargs.get('using')
        form_field.widget = ContentTypeAutocompleteSelect(
            db_field.remote_field,
            self.admin_site,
            using=db,
            attrs=attrs
        )

        return form_field

    def _get_gfk_ids_related_models(self, gfk):
        related_models = gfk.get_related_models()

        content_types = ContentType.objects.get_for_models(*related_models)

        return [x.id for x in content_types.values()]

    def _set_declared_gfk_to_form(self, form):
        for field in self._get_generic_foreign_keys(self.model):
            form_char_field = forms.CharField(
                widget=forms.HiddenInput(),
                required=False,
            )

            form.declared_fields[field.name] = form_char_field

    def _is_generic_foreign_key(self, field_name, generic_foreign_keys):
        for gfk in generic_foreign_keys:
            if field_name in (gfk.name, gfk.ct_field, gfk.fk_field):
                return True, gfk
        return False, None

    def _get_content_type_autocomplete_attrs(self, gfk):
        return {
            'class': 'generic-foreign-key-field',
            'data-ct--field': gfk.ct_field,
            'data-fk--field': gfk.fk_field,
            'data-gf--name': gfk.name,
            'data-gfk--models': ','.join(
                str(x) for x in self._get_gfk_ids_related_models(gfk)
            )
        }


class GenericForeignKeyInlineAdminMixin(BaseGenericForeignKeyMixin):
    def get_formset(self, request, obj, **kwargs):
        formset = super().get_formset(request, obj, **kwargs)
        self._set_declared_gfk_to_form(formset.form)
        self._set_validators_gfk(formset.form)

        return formset


class GenericForeignKeyAdminMixin(BaseGenericForeignKeyMixin):
    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj=obj)

        if fieldsets:
            fieldsets = self._update_gfk_fieldsets(fieldsets)

        return fieldsets

    def get_form(self, request, obj=None, change=False, **kwargs):
        self._set_declared_gfk_to_form(self.form)
        form = super().get_form(request, obj=obj, change=change, **kwargs)
        self._set_present_gfk(form, obj)
        self._set_validators_gfk(form)

        return form

    def _set_present_gfk(self, form, obj):
        for field in self._get_generic_foreign_keys(self.model):
            if field.fk_field in form.base_fields:
                form.base_fields[field.fk_field].widget.attrs[
                    'data-present'] = str(getattr(obj, field.name, ''))

    def get_inline_formsets(self, request, formsets, inline_instances,
                            obj=None):
        self.set_present_for_inlines(request, formsets, inline_instances, obj)

        return super().get_inline_formsets(
            request, formsets, inline_instances, obj=obj
        )

    def set_present_for_inlines(self, request, formsets, inline_instances, obj):
        for formset, inline in zip(formsets, inline_instances):
            generic_foreign_keys = self._get_generic_foreign_keys(formset.model)

            if not generic_foreign_keys:
                continue

            inline_fields = inline.get_fields(request, obj)

            for form in formset.forms:
                for gfk in generic_foreign_keys:
                    if gfk.ct_field not in inline_fields:
                        continue

                    form.fields[gfk.fk_field].widget.attrs[
                        'data-present'] = str(
                        getattr(form.instance, gfk.name, ''))

    def _update_gfk_fieldsets(self, fieldsets):
        gfk_fields = self._get_generic_foreign_keys(self.model)

        if not gfk_fields:
            return fieldsets

        for fieldset in fieldsets:
            fields = fieldset[1].get('fields')
            if fields:
                fieldset[1]['fields'] = self._update_gfk_fields(fields)
        return fieldsets



from django import forms

from admin_tools.dashboard.models import DashboardPreferences


class DashboardPreferencesForm(forms.ModelForm):
    """
    This form allows the user to edit dashboard preferences. It doesn't show
    the user field. It expects the user to be passed in from the view.
    """

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user', None)
        self.dashboard_id = kwargs.pop('dashboard_id', None)
        super(DashboardPreferencesForm, self).__init__(*args, **kwargs)

    def save(self, *args, **kwargs):
        preferences = super(DashboardPreferencesForm, self).save(
            commit=False,
            *args,
            **kwargs
        )
        preferences.user = self.user
        preferences.dashboard_id = self.dashboard_id
        preferences.save()
        return preferences

    class Meta:
        fields = ('data',)
        model = DashboardPreferences



import functools
import types


def copy_func(f):
    g = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__
    )
    g = functools.update_wrapper(g, f)
    return g



from django.core.checks import Warning, register
from django.template.loader import TemplateDoesNotExist, get_template

W001 = Warning(
    'You must add "admin_tools.template_loaders.Loader" in your '
    'template loaders variable, see: '
    'https://django-admin-tools.readthedocs.io/en/latest/configuration.html',
    id='admin_tools.W001',
    obj='admin_tools'
)


@register('admin_tools')
def check_admin_tools_configuration(app_configs=None, **kwargs):
    result = []
    try:
        get_template('admin:admin/base.html')
    except TemplateDoesNotExist:
        result.append(W001)
    return result



from tempfile import mktemp
from unittest import TestCase
from django.test import TestCase as DjangoTestCase
from django.core import management
from django.contrib.auth import models as auth_models

from admin_tools.dashboard import AppIndexDashboard
from admin_tools.dashboard.modules import DashboardModule, Group


class ManagementCommandTest(DjangoTestCase):
    def test_customdashboard(self):
        # check that customdashboard command doesn't raise exceptions
        file_name = mktemp()
        management.call_command('customdashboard', file=file_name)
        # and fails if file is already here
        try:
            management.call_command('customdashboard', file=file_name)
            assert False
        except:
            pass


class AppIndexDashboardTest(TestCase):
    def test_models(self):
        models = ['django.contrib.auth.models.User',
                  'django.contrib.auth.models.Group']
        board = AppIndexDashboard('Auth', models)
        self.assertEqual(board.get_app_model_classes(),
                         [auth_models.User, auth_models.Group])

__test__ = {
    'DashboardModule.is_empty': DashboardModule.is_empty,
    'DashboardModule.render_css_classes': DashboardModule.render_css_classes,
    'Group.is_empty': Group.is_empty,
}



from django.conf.urls import url
from admin_tools.dashboard import views

urlpatterns = [
    url(
        r'^set_preferences/(?P<dashboard_id>.+)/$',
        views.set_preferences,
        name='admin-tools-dashboard-set-preferences'
    ),
]



"""
Theming template tags.

To load the theming tags just do: ``{% load theming_tags %}``.
"""

from django import template
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils.safestring import mark_safe

register = template.Library()


def render_theming_css():
    """
    Template tag that renders the needed css files for the theming app.
    """
    css = getattr(settings, 'ADMIN_TOOLS_THEMING_CSS', False)
    if not css:
        css = '/'.join(['admin_tools', 'css', 'theming.css'])
    return mark_safe(
        '<link rel="stylesheet" type="text/css" media="screen" href="%s" />' %
        staticfiles_storage.url(css)
    )
register.simple_tag(render_theming_css)



from django.contrib.admin.widgets import AutocompleteSelect
from django import forms
from django.contrib import admin
from django.conf import settings
from django.contrib.admin.widgets import SELECT2_TRANSLATIONS, get_language
from django.forms.widgets import Media, MEDIA_TYPES


class AutocompleteFilter(admin.SimpleListFilter):
    template = 'admin_tools/autocomplete_filter/autocomplete_filter.html'
    title = ''
    field_name = ''
    is_placeholder_title = False
    widget_attrs = {}
    rel_model = None

    class Media:
        extra = '' if settings.DEBUG else '.min'
        i18n_name = SELECT2_TRANSLATIONS.get(get_language())
        i18n_file = (
        'admin/js/vendor/select2/i18n/%s.js' % i18n_name,) if i18n_name else ()
        js = (
                 'admin/js/vendor/jquery/jquery%s.js' % extra,
                 'admin/js/vendor/select2/select2.full%s.js' % extra,
             ) + i18n_file + (
                 'admin/js/jquery.init.js',
                 'admin/js/autocomplete.js',
                 'admin_tools/js/autocomplete_filter/autocomplete_filter_qs.js',
             )
        css = {
            'screen': (
                'admin/css/vendor/select2/select2%s.css' % extra,
                'admin/css/autocomplete.css',
                'admin_tools/css/autocomplete_filter/autocomplete_fix.css',
            ),
        }

    def __init__(self, request, params, model, model_admin):
        if self.parameter_name:
            raise AttributeError(
                'Rename attribute `parameter_name` to '
                '`field_name` for {}'.format(self.__class__)
            )
        self.parameter_name = '{}__id__exact'.format(self.field_name)
        super().__init__(request, params, model, model_admin)

        if self.rel_model:
            model = self.rel_model

        remote_field = model._meta.get_field(self.field_name).remote_field

        attrs = self.widget_attrs.copy()

        field = forms.ModelChoiceField(
            queryset=getattr(model, self.field_name).get_queryset(),
            widget=AutocompleteSelect(remote_field, model_admin.admin_site),
            required=False
        )

        self._add_media(model_admin)

        attrs['id'] = 'id-%s-dal-filter' % self.field_name

        if self.is_placeholder_title:
            attrs['data-placeholder'] = "By " + self.title

        self.rendered_widget = field.widget.render(
            name=self.parameter_name,
            value=self.used_parameters.get(self.parameter_name, ''),
            attrs=attrs
        )

    def _add_media(self, model_admin):
        if not hasattr(model_admin, 'Media'):
            raise Exception('Please add empty Media class to %s.' % model_admin)

        def _get_media(obj):
            return Media(media=getattr(obj, 'Media', None))

        media = _get_media(model_admin) + _get_media(
            AutocompleteFilter) + _get_media(self)

        for name in MEDIA_TYPES:
            setattr(model_admin.Media, name, getattr(media, "_" + name))

    def has_output(self):
        return True

    def lookups(self, request, model_admin):
        return ()

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(**{self.parameter_name: self.value()})
        else:
            return queryset



from admin_tools.services.urls import admin_autocomplete_url
from django.contrib.admin.views.autocomplete import AutocompleteJsonView
from django.http import Http404, JsonResponse
from django.utils.text import capfirst


class ContentTypeAutocompleteView(AutocompleteJsonView):
    # copied from base class
    def get(self, request, *args, **kwargs):
        if not self.model_admin.get_search_fields(request):
            raise Http404(
                '%s must have search_fields for the autocomplete_view.' %
                type(self.model_admin).__name__
            )
        if not self.has_perm(request):
            return JsonResponse({'error': '403 Forbidden'}, status=403)

        self.term = request.GET.get('term', '')
        self.paginator_class = self.model_admin.paginator
        self.object_list = self.get_queryset()
        context = self.get_context_data()
        return JsonResponse({
            'results': [
                {
                    'id': str(obj.pk),
                    'text': capfirst(obj.model),
                    # modify
                    'autocompleteUrl': self.get_autocomplete_ul(obj)
                }
                for obj in context['object_list']
            ],
            'pagination': {'more': context['page_obj'].has_next()},
        })

    def get_queryset(self):
        queryset = super().get_queryset()

        ids = self.request.GET.get('ids')

        if ids:
            queryset = queryset.filter(
                id__in=[int(x) for x in ids.split(',')]
            )

        return queryset

    def get_autocomplete_ul(self, content_type):
        return admin_autocomplete_url(content_type.model_class())



from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse
from django.shortcuts import render
from django.contrib import messages

try:
    from django.views.decorators.csrf import csrf_exempt
except ImportError:
    from django.contrib.csrf.middleware import csrf_exempt

from .forms import DashboardPreferencesForm
from .models import DashboardPreferences


@staff_member_required
@csrf_exempt
def set_preferences(request, dashboard_id):
    """
    This view serves and validates a preferences form.
    """
    try:
        preferences = DashboardPreferences.objects.get(
            user=request.user,
            dashboard_id=dashboard_id
        )
    except DashboardPreferences.DoesNotExist:
        preferences = None
    if request.method == "POST":
        form = DashboardPreferencesForm(
            user=request.user,
            dashboard_id=dashboard_id,
            data=request.POST,
            instance=preferences
        )
        if form.is_valid():
            preferences = form.save()
            if request.is_ajax():
                return HttpResponse('true')
            messages.success(request, 'Preferences saved')
        elif request.is_ajax():
            return HttpResponse('false')
    else:
        form = DashboardPreferencesForm(
            user=request.user,
            dashboard_id=dashboard_id,
            instance=preferences
        )
    return render(
        request,
        'admin_tools/dashboard/preferences_form.html',
         {'form': form}
    )



from django import forms
from django.contrib.admin.widgets import (
    AutocompleteSelect as BaseAutocompleteSelect
)
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse


class ContentTypeAutocompleteSelect(BaseAutocompleteSelect):
    @property
    def media(self):
        media = super().media

        media_js = media._js if media else []
        media_css = media._css if media else {}

        return forms.Media(
            js=tuple([*media_js,
                      'admin_tools/js/widgets/generic-foreign-key-field.js']),
            css=media_css
        )

    def create_option(
        self, name, value, label, selected, index, subindex=None, attrs=None
    ):
        result = super().create_option(
            name, value, label, selected, index, subindex=subindex, attrs=attrs
        )

        if value:
            url = self.get_autocomplete_url(value)
            result['attrs']['data-autocomplete-url'] = url

        return result

    def get_autocomplete_url(self, ct_id):
        url_name = 'admin:%s_%s_autocomplete'
        meta = ContentType.objects.get(id=ct_id).model_class()._meta

        return reverse(url_name % (meta.app_label, meta.model_name))






from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.contrib import messages

try:
    from django.views.decorators.csrf import csrf_exempt
except ImportError:
    from django.contrib.csrf.middleware import csrf_exempt

from .forms import BookmarkForm
from .models import Bookmark


@staff_member_required
@csrf_exempt
def add_bookmark(request):
    """
    This view serves and validates a bookmark form.
    If requested via ajax it also returns the drop bookmark form to replace the
    add bookmark form.
    """
    if request.method == "POST":
        form = BookmarkForm(user=request.user, data=request.POST)
        if form.is_valid():
            bookmark = form.save()
            if not request.is_ajax():
                messages.success(request, 'Bookmark added')
                if request.POST.get('next'):
                    return HttpResponseRedirect(request.POST.get('next'))
                return HttpResponse('Added')
            return render(
                request,
                'admin_tools/menu/remove_bookmark_form.html',
                {'bookmark': bookmark, 'url': bookmark.url}
            )
    else:
        form = BookmarkForm(user=request.user)
    return render(
        request,
        'admin_tools/menu/form.html',
        {'form': form, 'title': 'Add Bookmark'}
    )


@staff_member_required
@csrf_exempt
def edit_bookmark(request, id):
    bookmark = get_object_or_404(Bookmark, id=id)
    if request.method == "POST":
        form = BookmarkForm(
            user=request.user,
            data=request.POST,
            instance=bookmark
        )
        if form.is_valid():
            form.save()
            if not request.is_ajax():
                messages.success(request, 'Bookmark updated')
                if request.POST.get('next'):
                    return HttpResponseRedirect(request.POST.get('next'))
            return HttpResponse('Saved')
    else:
        form = BookmarkForm(user=request.user, instance=bookmark)
    return render_to_response(
        'admin_tools/menu/form.html',
        {'form': form, 'title': 'Edit Bookmark'}
    )


@staff_member_required
@csrf_exempt
def remove_bookmark(request, id):
    """
    This view deletes a bookmark.
    If requested via ajax it also returns the add bookmark form to replace the
    drop bookmark form.
    """
    bookmark = get_object_or_404(Bookmark, id=id, user=request.user)
    if request.method == "POST":
        bookmark.delete()
        if not request.is_ajax():
            messages.success(request, 'Bookmark removed')
            if request.POST.get('next'):
                return HttpResponseRedirect(request.POST.get('next'))
            return HttpResponse('Deleted')
        return render(
            request,
            'admin_tools/menu/add_bookmark_form.html',
            {
                'url': request.POST.get('next'),
                'title': '**title**'  # replaced on the javascript side
            }
        )
    return render(
        request,
        'admin_tools/menu/delete_confirm.html',
        {'bookmark': bookmark, 'title': 'Delete Bookmark'}
    )



from contextlib import suppress

from django.urls import reverse, NoReverseMatch


def admin_change_url(obj):
    app_label = obj._meta.app_label
    model_name = obj._meta.model.__name__.lower()
    return reverse(f'admin:{app_label}_{model_name}_change', args=(obj.pk,))


def admin_changelist_url(model):
    app_label = model._meta.app_label
    model_name = model.__name__.lower()
    return reverse(f'admin:{app_label}_{model_name}_changelist')


def admin_autocomplete_url(model):
    app_label = model._meta.app_label
    model_name = model.__name__.lower()
    with suppress(NoReverseMatch):
        return reverse(f'admin:{app_label}_{model_name}_autocomplete')



from django.conf.urls import url
from admin_tools.menu import views

urlpatterns = [
    url(
        r'^add_bookmark/$',
        views.add_bookmark,
        name='admin-tools-menu-add-bookmark'
    ),
    url(
        r'^edit_bookmark/(?P<id>.+)/$',
        views.edit_bookmark,
        name='admin-tools-menu-edit-bookmark'
    ),
    url(
        r'^remove_bookmark/(?P<id>.+)/$',
        views.remove_bookmark,
        name='admin-tools-menu-remove-bookmark'
    ),
]



from django.apps import AppConfig


class AdminToolsConfig(AppConfig):
    name = 'admin_tools'

    def ready(self):
        super().ready()
        # load admin_tools checks
        from . import checks # noqa: F401
